open Arg;;
open Utils;;
open List;;
open Str;;

open Printf;;

let user = ref false;;
let spec = [("-user", Unit (fun () -> user := true), " Show only user time")];;
let msg = "Parser for benchmark files.";;
let files = ref [];;

(* vmname * date * (testname * (key, value) list) list *)
type entry = string * (string * float) list
type test  = string * string * entry list

let read_lines file =
  Str.split (regexp "\n") (read_file file (20*80))
;;

let rec assoc_list = function
  | [] -> []
  | a::b::xs -> (a,b) :: assoc_list xs
  | _ -> raise (Failure "associative lists must be of even length: ")
;;

let dokeyval s = match split (regexp ":") s with
  | [k;v] -> (k, float_of_string v)
  | _ -> raise (Failure "invalid key value pair")
;;

let doline l = match split (regexp "[ \t]+") l with
  | (testname :: keyvals) -> (testname, map dokeyval keyvals)
  | _ -> raise (Failure "invalid line")
;;

let dohead s = match split (regexp " - ") s with
  | [vm; date] -> (vm, date)
  | _ -> raise (Failure "invalid header")
;;
  
let dofile f : test = match read_lines f with
  | header :: lines ->
      let (vm, date) = dohead header in
	(vm, date, map doline lines)
  | _ -> raise (Failure ("invalid file: " ^ f))
;;

let maxstr s l =
  if String.length s > l
  then String.sub s 0 l
  else s
;;

let rec print_header = function
  | [] -> printf "\n"
  | (h,_d,_es)::ts ->
      printf "%-14s" (maxstr h 10);
      print_header ts
;;

let tr_list (xys : 'a list list) : 'a list * 'a list list =
  fold_right
    (function
       | x::[] -> (fun (xs, ys) -> (x::xs, []))
       | x::y  -> (fun (xs, ys) -> (x::xs, y::ys)))
    xys
    ([],[])
;;

let print_keyval (i:int) kv =
  if i < length kv
  then let (k,v) = nth kv i in
    (if not !user || String.compare k "user" = 0
     then printf " %5s %6.2f " k v)
  else
    printf "%11s " "----"
;;

let print_entries es =
  let (tname, _) = hd es in
  let kvs = map (fun (_,kv) -> kv) es in
  let len = fold_left (fun m l -> max m (length l)) 0 kvs in
  let s = ref (maxstr tname 10) in
    if not !user
    then (for i = 0 to len-1 do
	    printf "%10s" !s; s:="";
	    (iter (print_keyval i) kvs);
	    printf "\n"
	  done)
    else (printf "%10s" !s;
	  for i = 0 to len-1 do
	    (iter (print_keyval i) kvs);
	  done;
	  printf "\n")
;;

let rec print_tests es =
  match tr_list es with
    | (xs,[]) -> ()
    | (xs,ys) ->
	print_entries xs;
	print_tests ys
;;

let main () =
  parse spec (fun s -> files := s :: !files) msg;
  let tests = map dofile (rev !files) in
  let fno = length !files in
  let _tno = length tests in
    (if fno < 1 then raise (Failure "please supplied some files."));
    printf "%12s" "";
    print_header tests;
    print_tests (map (fun (_,_,es) -> es) tests);
;;

let _ = main ();;
