#include "defines.h"

exception PARSE_ERROR of string;;
exception RUNTIME_ERROR of string;;
exception TYPE_ERROR of string;;
exception UNSUPPORTED_ERROR of string;;
exception INTERNAL_ERROR of string;;
exception ASSERTION_ERROR of string;;

let error_prefix =
  "ERROR: "
;;

let parse_error s =
  raise (PARSE_ERROR (error_prefix ^ s))
;;

let runtime_error s =
  raise (RUNTIME_ERROR (error_prefix ^ s))
;;

let internal_error s =
  raise (INTERNAL_ERROR (error_prefix ^ s))
;;

let unsupported_error s =
  raise (UNSUPPORTED_ERROR (error_prefix ^ "unsupported feature: " ^ s))
;;

let unsigned_char = function
  | c when c < 0 || c > 255 ->
      parse_error ("out of character range ("^string_of_int c^")")
  | c when c > 127 -> c - 256
  | c -> c
;;


let id x = x;;

let maybe d = function
  | None -> d
  | Some v -> v
;;

(* Timing *)
let gettime () =
  Unix.gettimeofday ()
;;

(* Simple read file method *)
let read_file file exp_len =
  let buf = Buffer.create exp_len in
  let ic  = open_in file in
  let rec loop () =
    Buffer.add_char buf (input_char ic);
    loop ()
  in
    try
      loop ()
    with End_of_file ->
      close_in ic;
      Buffer.contents buf
;;

(* Pretty Printing Utilities (c) John Harrop
 * http:/ /alan.petitepomme.net/cwn/2008.07.08.html#4 *)
include Format;;
let rec fprintfs sep f ff = function
  | [] -> ()
  | [h] -> fprintf ff "%a" f h
  | h::t -> fprintf ff "%a%a%a" f h sep () (fprintfs sep f) t;;

let sep_space   f = fprintfs (fun ff () -> fprintf ff "@ ") f;;
let sep_colon   f = fprintfs (fun ff () -> fprintf ff ":") f;;
let sep_semi    f = fprintfs (fun ff () -> fprintf ff ";") f;;
let sep_comma   f = fprintfs (fun ff () -> fprintf ff ",") f;;
let sep_period  f = fprintfs (fun ff () -> fprintf ff ".") f;;
let sep_newline f = fprintfs (fun ff () -> fprintf ff "@\n") f;;
(* strict variants *)
let ssep_space   f = fprintfs (fun ff () -> fprintf ff " ") f;;
let ssep_colon   f = fprintfs (fun ff () -> fprintf ff ":") f;;
let ssep_semi    f = fprintfs (fun ff () -> fprintf ff ";") f;;
let ssep_comma   f = fprintfs (fun ff () -> fprintf ff ",") f;;
let ssep_period  f = fprintfs (fun ff () -> fprintf ff ".") f;;
let ssep_newline f = fprintfs (fun ff () -> fprintf ff "\n") f;;

let fprintf_int ff = fprintf ff "%d";;
let fprintf_string ff = fprintf ff "%s";;
let fprintf_list f ff list =
  fprintf ff "[@[%a@]]" (sep_semi f) list;;
let fprintf_array f ff array =
  fprintf ff "[|@[%a@]|]" (sep_semi f) (Array.to_list array);;
let fprintf_option f ff = function
  | None -> fprintf ff "None"
  | Some x -> fprintf ff "Some %a" f x;;
